searchNodes=[{"doc":"A pool for concurrent transactional tests. The sandbox pool is implemented on top of an ownership mechanism. When started, the pool is in automatic mode, which means the repository will automatically check connections out as with any other pool. The mode/2 function can be used to change the pool mode to manual or shared. In both modes, the connection must be explicitly checked out before use. When explicit checkouts are made, the sandbox will wrap the connection in a transaction by default and control who has access to it. This means developers have a safe mechanism for running concurrent tests against the database. Database support While this sandbox has been developped with CockroachDB in mind, it should work with other Postgresql variants. Example The first step is to configure your database to use the EctoReplaySandbox pool in your tests. You set this in your config/test.exs if you haven't yet: config :my_app , MyApp.Repo , pool : EctoReplaySandbox Now with the test database properly configured, you can write transactional tests: # At the end of your test_helper.exs # Set the pool mode to manual for explicit checkouts sandbox = Application . get_env ( :my_app , Repo ) [ :pool ] sandbox . mode ( Repo , :manual ) defmodule PostTest do # Once the mode is manual, tests can also be async use ExUnit.Case , async : true setup do # Explicitly get a connection before each test sandbox = Application . get_env ( :my_app , Repo ) [ :pool ] :ok = sandbox . checkout ( Repo ) end test &quot;create post&quot; do # Use the repository as usual assert % Post { } = Repo . insert! ( % Post { } ) end end Collaborating processes The example above is straight-forward because we have only a single process using the database connection. However, sometimes a test may need to interact with multiple processes, all using the same connection so they all belong to the same transaction. Before we discuss solutions, let's see what happens if we try to use a connection from a new process without explicitly checking it out first: setup do # Explicitly get a connection before each test sandbox = Application . get_env ( :my_app , Repo ) [ :pool ] :ok = sandbox . checkout ( Repo ) end test &quot;create two posts, one sync, another async&quot; do task = Task . async ( fn -&gt; Repo . insert! ( % Post { title : &quot;async&quot; } ) end ) assert % Post { } = Repo . insert! ( % Post { title : &quot;sync&quot; } ) assert % Post { } = Task . await ( task ) end The test above will fail with an error similar to: ** (RuntimeError) cannot find ownership process for #PID&lt;0.35.0&gt; That's because the setup block is checking out the connection only for the test process. Once we spawn a Task, there is no connection assigned to it and it will fail. The sandbox module provides two ways of doing so, via allowances or by running in shared mode. Allowances The idea behind allowances is that you can explicitly tell a process which checked out connection it should use, allowing multiple processes to collaborate over the same connection. Let's give it a try: test &quot;create two posts, one sync, another async&quot; do parent = self ( ) task = Task . async ( fn -&gt; sandbox = Application . get_env ( :my_app , Repo ) [ :pool ] sandbox . allow ( Repo , parent , self ( ) ) Repo . insert! ( % Post { title : &quot;async&quot; } ) end ) assert % Post { } = Repo . insert! ( % Post { title : &quot;sync&quot; } ) assert % Post { } = Task . await ( task ) end And that's it, by calling allow/3 , we are explicitly assigning the parent's connection (i.e. the test process' connection) to the task. Because allowances use an explicit mechanism, their advantage is that you can still run your tests in async mode. The downside is that you need to explicitly control and allow every single process. This is not always possible. In such cases, you will want to use shared mode. Shared mode Shared mode allows a process to share its connection with any other process automatically, without relying on explicit allowances. Let's change the example above to use shared mode: setup do # Explicitly get a connection before each test sandbox = Application . get_env ( :my_app , Repo ) [ :pool ] :ok = sandbox . checkout ( Repo ) # Setting the shared mode must be done only after checkout sandbox . mode ( Repo , { :shared , self ( ) } ) end test &quot;create two posts, one sync, another async&quot; do task = Task . async ( fn -&gt; Repo . insert! ( % Post { title : &quot;async&quot; } ) end ) assert % Post { } = Repo . insert! ( % Post { title : &quot;sync&quot; } ) assert % Post { } = Task . await ( task ) end By calling mode({:shared, self()}) , any process that needs to talk to the database will now use the same connection as the one checked out by the test process during the setup block. Make sure to always check a connection out before setting the mode to {:shared, self()} . The advantage of shared mode is that by calling a single function, you will ensure all upcoming processes and operations will use that shared connection, without a need to explicitly allow them. The downside is that tests can no longer run concurrently in shared mode. Summing up There are two mechanisms for explicit ownerships: Using allowances - requires explicit allowances via allow/3 . Tests may run concurrently. Using shared mode - does not require explicit allowances. Tests cannot run concurrently. FAQ When running the sandbox mode concurrently, developers may run into issues we explore in the upcoming sections. &quot;owner exited while client is still running&quot; In some situations, you may see error reports similar to the one below: 21 : 57 : 43.910 [ error ] Postgrex.Protocol ( # PID &lt; 0 . 284 . 0 &gt; ) disconnected : ** (DBConnection.Error) owner #PID&lt;&gt; exited while client #PID&lt;&gt; is still running Such errors are usually followed by another error report from another process that failed while executing a database query. To understand the failure, we need to answer the question: who are the owner and client processes? The owner process is the one that checks out the connection, which, in the majority of cases, is the test process, the one running your tests. In other words, the error happens because the test process has finished, either because the test succeeded or because it failed, while the client process was trying to get information from the database. Since the owner process, the one that owns the connection, no longer exists, Ecto will check the connection back in and notify the client process using the connection that the connection owner is no longer available. This can happen in different situations. For example, imagine you query a GenServer in your test that is using a database connection: test &quot;gets results from GenServer&quot; do { :ok , pid } = MyAppServer . start_link ( ) sandbox = Application . get_env ( :my_app , Repo ) [ :pool ] sandbox . allow ( Repo , self ( ) , pid ) assert MyAppServer . get_my_data_fast ( timeout : 1000 ) == [ ... ] end In the test above, we spawn the server and allow it to perform database queries using the connection owned by the test process. Since we gave a timeout of 1 second, in case the database takes longer than one second to reply, the test process will fail, due to the timeout, making the &quot;owner down&quot; message to be printed because the server process is still waiting on a connection reply. In some situations, such failures may be intermittent. Imagine that you allow a process that queries the database every half second: test &quot;queries periodically&quot; do { :ok , pid } = PeriodicServer . start_link ( ) sandbox = Application . get_env ( :my_app , Repo ) [ :pool ] sandbox . allow ( Repo , self ( ) , pid ) # more tests end Because the server is querying the database from time to time, there is a chance that, when the test exists, the periodic process may be querying the database, regardless of test success or failure. &quot;owner timed out because it owned the connection for longer than Nms&quot; In some situations, you may see error reports similar to the one below: 09 : 56 : 43.081 [ error ] Postgrex.Protocol ( # PID &lt; &gt; ) disconnected : ** (DBConnection.ConnectionError) owner #PID&lt;&gt; timed out because it owned the connection for longer than 15000ms If you have a long running test (or you're debugging with IEx.pry), the timeout for the connection ownership may be too short. You can increase the timeout by setting the :ownership_timeout options for your repo config in config/config.exs (or preferably in config/test.exs ): config :my_app , MyApp.Repo , ownership_timeout : NEW_TIMEOUT_IN_MILLISECONDS The :ownership_timeout option is part of DBConnection.Ownership and defaults to 15000ms. Timeouts are given as integers in milliseconds. Alternately, if this is an issue for only a handful of long-running tests, you can pass an :ownership_timeout option when calling EctoReplaySandbox.checkout/2 instead of setting a longer timeout globally in your config. Database locks and deadlocks Since the sandbox relies on concurrent transactional tests, there is a chance your tests may trigger deadlocks in your database. This is specially true with MySQL, where the solutions presented here are not enough to avoid deadlocks and therefore making the use of concurrent tests with MySQL prohibited. However, even on databases like PostgreSQL, performance degradations or deadlocks may still occur. For example, imagine multiple tests are trying to insert the same user to the database. They will attempt to retrieve the same database lock, causing only one test to succeed and run while all other tests wait for the lock. In other situations, two different tests may proceed in a way that each test retrieves locks desired by the other, leading to a situation that cannot be resolved, a deadlock. For instance: Transaction 1 : Transaction 2 : begin begin update posts where id = 1 update posts where id = 2 update posts where id = 1 update posts where id = 2 * * deadlock * * There are different ways to avoid such problems. One of them is to make sure your tests work on distinct data. Regardless of your choice between using fixtures or factories for test data, make sure you get a new set of data per test. This is specially important for data that is meant to be unique like user emails. For example, instead of: def insert_user do Repo . insert! % User { email : &quot;sample@example.com&quot; } end prefer: def insert_user do Repo . insert! % User { email : &quot;sample- \#{ counter ( ) } @example.com&quot; } end defp counter do System . unique_integer [ :positive ] end Deadlocks may happen in other circumstances. If you believe you are hitting a scenario that has not been described here, please report an issue so we can improve our examples. As a last resort, you can always disable the test triggering the deadlock from running asynchronously by setting &quot;async: false&quot;.","ref":"EctoReplaySandbox.html","title":"EctoReplaySandbox","type":"module"},{"doc":"Allows the allow process to use the same connection as parent .","ref":"EctoReplaySandbox.html#allow/4","title":"EctoReplaySandbox.allow/4","type":"function"},{"doc":"Checks in the connection back into the sandbox pool.","ref":"EctoReplaySandbox.html#checkin/2","title":"EctoReplaySandbox.checkin/2","type":"function"},{"doc":"Checks a connection out for the given repo . The process calling checkout/2 will own the connection until it calls checkin/2 or until it crashes when then the connection will be automatically reclaimed by the pool. Options :sandbox - when true the connection is wrapped in a transaction. Defaults to true. :isolation - set the query to the given isolation level. :ownership_timeout - limits how long the connection can be owned. Defaults to the value in your repo config in config/config.exs (or preferably in config/test.exs ), or 60000 ms if not set. The timeout exists for sanity checking purposes, to ensure there is no connection leakage, and can be bumped whenever necessary.","ref":"EctoReplaySandbox.html#checkout/2","title":"EctoReplaySandbox.checkout/2","type":"function"},{"doc":"Sets the mode for the repo pool. The mode can be :auto , :manual or {:shared, &lt;pid&gt;} . Warning: you should only call this function in the setup block for a test and not within a test, because if the mode is changed during the test it will cause other database connections to be checked in (causing errors).","ref":"EctoReplaySandbox.html#mode/2","title":"EctoReplaySandbox.mode/2","type":"function"},{"doc":"Runs a function outside of the sandbox.","ref":"EctoReplaySandbox.html#unboxed_run/2","title":"EctoReplaySandbox.unboxed_run/2","type":"function"},{"doc":"This is a custom implementation of Ecto.Adapters.SQL.Sandbox designed to work with CockroachDB by not leveraging savepoints. Each test runs inside a transaction managed by the Sandbox, just like with default Ecto Sandbox. Inside your test, when your code opens a transaction block, given that CockroachDB does not support nested transactions or savepoints, no actual database transaction is created. Instead the sandbox is using a log approach described below and such transaction are called pseudo transaction. The sandbox maintains 2 logs for a given managed transaction: Sandbox log Transaction log The Sandbox log contains the statements that have been commited to the database by your test, either outside of a pseudo transaction or once a pseudo transaction is commited. The Transaction log keeps track of statements that are running inside a pseudo transaction and it is eventually appended to the Sandbox log once the pseudo transaction is successfully committed. When an error occurs or when a pseudo transaction is being explicitely rollbacked, the managed transaction is being rollbacked and a new transaction is being created. Then the Sandbox log is being replayed to restore the state to how it was before the error or the rollbacked pseudo transaction started. Once the test finishes, the managed transaction is being rollbacked to restore the state of the database to how it was before the test began.","ref":"readme.html","title":"EctoReplaySandbox","type":"extras"},{"doc":"The package can be installed by adding ecto_replay_sandbox to your list of dependencies in mix.exs . def deps do [ { :ecto_replay_sandbox , &quot;~&gt; 2.1&quot; , only : :test } , ] end","ref":"readme.html#installation","title":"EctoReplaySandbox - Installation","type":"extras"},{"doc":"In your config/test.ex config :my_app , MyApp.Repo , pool : EctoReplaySandbox In your test/test_helper.ex Replace the following line: Ecto.Adapters.SQL.Sandbox . mode ( MyApp.Repo , :manual ) with: sandbox = Application . get_env ( :my_app , MyApp.Repo ) [ :pool ] sandbox . mode ( MyApp.Repo , :manual ) Do the same for your test/support/xxx_case.ex files, for example: Replace the following line: setup tags do :ok = Ecto.Adapters.SQL.Sandbox . checkout ( MyApp.Repo ) unless tags [ :async ] do Ecto.Adapters.SQL.Sandbox . mode ( MyApp.Repo , :manual ) end { :ok , conn : Phoenix.ConnTest . build_conn ( ) } end with: setup tags do sandbox = Application . get_env ( :myapp , MyApp.Repo ) [ :pool ] :ok = sandbox . checkout ( MyApp.Repo ) unless tags [ :async ] do sandbox . mode ( MyApp.Repo , { :shared , self ( ) } ) end { :ok , conn : Phoenix.ConnTest . build_conn ( ) } end This effectively removes the hardcoded usage of Ecto.Adapters.SQL.Sandbox with a dynamic lookup of the configured pool.","ref":"readme.html#usage","title":"EctoReplaySandbox - Usage","type":"extras"},{"doc":"I'd like to give special thanks to James Fish (@fishcakez) for his help getting this working and always taking the time to offer guidance or otherwise feedback the CockroachDB team so that we can use Ecto with CockroachDB with little to no friction.","ref":"readme.html#credits","title":"EctoReplaySandbox - Credits","type":"extras"}]